import numpy as np
import matplotlib.pyplot as plt
from matplotlib import cm
from matplotlib.widgets import Button, Slider
import warnings
warnings.simplefilter('always')

color_cycle = plt.cm.tab10.colors[:10]

def eprplot(eprdata_list, plot_type='stacked', 
            slices='all', spacing=0.5, 
            plot_imag=True,g_scale=False, interactive=False):

    """
    Plot one or multiple EPR data objects for visualization and comparison.

    Parameters
    ----------
    eprdata_list : list of EprData
        A list of `EprData` objects to be plotted. Each object should have the following attributes:
        - `filepath` (str): Path to the data file.
        - `data` (ndarray): Data array, which can be 1D or 2D, real or complex.
        - `acq_param` (dict): Acquisition parameters.
    
    plot_type : {'stacked', 'superimposed', 'surf', 'contour'}, optional
        The type of plot to generate for 2D data. Options are:
        - 'stacked' (default): Stacks slices with specified spacing.
        - 'superimposed': Overlays all slices.
        - 'surf': Creates a 3D surface plot.
        - 'contour': Creates a 2D contour plot.

    slices : {'all', list, range}, optional
        Specifies which slices to plot for 2D data. Options include:
        - 'all' (default): Plots all slices.
        - A list of integers: Specifies the indices of slices to plot.
        - A range object: Specifies a range of slice indices.

    spacing : float, optional
        Spacing between slices for 'stacked' or 'superimposed' plots. Default is 0.5.

    plot_imag : bool, optional
        If `True`, includes the imaginary part of the data in the plot. Default is `True`.

    g_scale : bool, optional
        If `True` and `EprData.g` is not `None`, plots the data on a g-factor scale 
        and inverts the x-axis. Default is `False`.

    Returns
    -------
    fig : matplotlib.figure.Figure
        The figure object containing the plot.
    ax : matplotlib.axes._axes.Axes
        The axes object of the plot.

    Notes
    -----
    - For 1D data, all `EprData` objects in the list must have 1D data.
    - For 2D data, all `EprData` objects in the list must have 2D data.
    - The function automatically handles plotting based on the data dimensions 
      (1D or 2D) and the specified plot type.
    - If `g_scale` is enabled and `EprData.g` is available, the x-axis is inverted 
      for better visualization of g-factor values.
    """

    
    if not isinstance(eprdata_list, list):
        eprdata_list = [eprdata_list]  # Convert single object to a list for uniform handling
    
    if eprdata_list[0].data.ndim==1:
        assert all([i.data.ndim==1 for i in eprdata_list]), 'Only datasets with same number of dimensions can be compared.'
        ndim=1
    if eprdata_list[0].data.ndim==2:
        assert all([i.data.ndim==2 for i in eprdata_list]), 'Only datasets with same number of dimensions can be compared.'
        ndim=2

    if ndim==1:
        fig,ax = plot_1d(eprdata_list,g_scale,plot_imag)
    elif ndim==2:
        fig,ax = plot_2d(eprdata_list,g_scale,plot_type, slices, spacing)

    if plot_type not in ['slider','surf']:
        fig.set_size_inches((4,3))
        fig.tight_layout()
    
    if g_scale and eprdata_list[0].g is not None: 
        ax.invert_xaxis()
    
    if interactive:
        data_cursor(fig)
    
    return fig,ax

def plot_1d(eprdata_list,g_scale,plot_imag=True):

    """
    Plot 1D EPR data from multiple `EprData` objects.

    Parameters
    ----------
    eprdata_list : list of EprData
        A list of `EprData` objects to be plotted. Each object should have the following attributes:
        - `data` (ndarray): 1D data array (real or complex).
        - `x` (ndarray): x-axis values corresponding to the data.
        - `g` (ndarray, optional): g-factor values, if applicable.
    
    g_scale : bool
        If `True`, plots the data on a g-factor scale (using `g` if available). Default is `False`.

    plot_imag : bool, optional
        If `True`, includes the imaginary part of the data in the plot for complex data. Default is `True`.

    Returns
    -------
    fig : matplotlib.figure.Figure
        The figure object containing the plot.
    ax : matplotlib.axes._axes.Axes
        The axes object of the plot.

    Notes
    -----
    - If `g_scale` is enabled and the `g` attribute is present in the `EprData` object, 
      it will be used for the x-axis. Otherwise, the x-values (`x`) will be used.
    - Complex data is plotted with the real and imaginary parts (if `plot_imag` is `True`).
    - The function uses a cycling color scheme for each plot in `eprdata_list`.
    """
    
    c_idx = -1

    # initiate plot
    fig,ax = plt.subplots()

    for idx, eprdata in enumerate(eprdata_list):

        if c_idx>=9:
            c_idx=-1
        c_idx+=1

        data = eprdata.data
        if g_scale:
            if eprdata.g is not None:
                x = eprdata.g
            else:
                x = eprdata.x
                warnings.warn('Unable to set g values as axis.')  
        else:
            x = eprdata.x

        if np.iscomplexobj(data):
            ax.plot(x,np.real(data), label='Real',color=color_cycle[c_idx])
            if plot_imag:
                ax.plot(x,np.imag(data), '--', alpha=0.5, label='Imaginary',color=color_cycle[c_idx])
        else:
            ax.plot(x,data, label='Real',color=color_cycle[c_idx])
    
    return fig,ax
        

def plot_2d(eprdata_list,g_scale,plot_type='stacked',slices='all', spacing=0.5):

    """
    Plot 2D EPR data from multiple `EprData` objects.

    Parameters
    ----------
    eprdata_list : list of EprData
        A list of `EprData` objects to be plotted. Each object should have the following attributes:
        - `data` (ndarray): 2D data array.
        - `x` (ndarray): x-axis values corresponding to the data.
        - `y` (ndarray): y-axis values corresponding to the data.
        - `g` (ndarray, optional): g-factor values, if applicable.
    
    g_scale : bool
        If `True`, plots the data on a g-factor scale (using `g` if available). Default is `False`.

    plot_type : {'stacked', 'superimposed', 'surf', 'pcolor'}, optional
        The type of plot to generate for the 2D data. Options are:
        - 'stacked' (default): Stacks slices with specified spacing.
        - 'superimposed': Overlays slices on the same plot.
        - 'surf': Creates a 3D surface plot.
        - 'pcolor': Creates a pseudocolor plot.

    slices : {'all', list, range}, optional
        Specifies which slices to plot for the 2D data. Options include:
        - 'all' (default): Plots all slices.
        - A list of integers: Specifies the indices of slices to plot.
        - A range object: Specifies a range of slice indices.

    spacing : float, optional
        Spacing between slices for 'stacked' and 'superimposed' plots. Default is 0.5.

    Returns
    -------
    fig : matplotlib.figure.Figure
        The figure object containing the plot.
    ax : matplotlib.axes._axes.Axes
        The axes object of the plot.

    Notes
    -----
    - For 2D data, the function selects slices based on the `slices` parameter and 
      visualizes them using one of the specified plot types.
    - The color scheme for the slices is determined by a gradual color map.
    - If `g_scale` is enabled and the `g` attribute is available in the `EprData` object, 
      it will be used for the x-axis. Otherwise, the x-values (`x`) will be used.
    """

    data = eprdata_list[0].data
    if g_scale:
        if eprdata_list[0].g is not None:
            x = eprdata_list[0].g
        else:
            x = eprdata_list[0].x
            warnings.warn('Unable to set g values as axis.')  
    else:
        x = eprdata_list[0].x
    y = eprdata_list[0].y
    num_slices, slice_len = data.shape

    # Select slices
    if slices == 'all':
        selected_slices = range(num_slices)
    elif isinstance(slices, list):
        selected_slices = [s for s in slices if 0 <= s < num_slices]
    elif isinstance(slices, range):
        selected_slices = [s for s in slices if 0 <= s < num_slices]
    else:
        raise ValueError("Invalid value for 'slices'. Must be 'all', a list, or a range.")

    # Set color scheme for slices
    num_selected_slices = len(selected_slices)

    slice_colors = cm.winter(np.linspace(0, 1, num_selected_slices)) 

    if plot_type=='surf':
        fig,ax = surf_plot(data,x,y,slice_len,selected_slices)
    elif plot_type=='superimposed':
        fig,ax = superimposed_plot(data,x,y,selected_slices,slice_colors)
    elif plot_type=='stacked':
        fig,ax = stack_plot(data,x,y,selected_slices,slice_colors,spacing)
    elif plot_type=='pcolor':
        fig,ax = pcolor_plot(data,x,y,slice_len,selected_slices)
    elif plot_type=='slider':
        fig,ax = slider_plot(data,x,y,slice_len,selected_slices)

    return fig,ax

def stack_plot(data,x,y,selected_slices,slice_colors,spacing):

    """
    Create a stacked plot for 2D EPR data slices.

    Parameters
    ----------
    data : ndarray
        The 2D data array from which the slices will be plotted.
    x : ndarray
        The x-axis values corresponding to the data.
    y : ndarray
        The y-axis values corresponding to the data.
    selected_slices : list of int
        A list of slice indices to be plotted.
    slice_colors : ndarray
        An array of colors to be used for each slice.
    spacing : float
        The vertical spacing between stacked slices.

    Returns
    -------
    fig : matplotlib.figure.Figure
        The figure object containing the stacked plot.
    ax : matplotlib.axes._axes.Axes
        The axes object of the plot.

    Notes
    -----
    - This function plots each slice of the 2D data array, offset vertically by the 
      specified `spacing` to create a stacked appearance.
    - The slices are drawn with colors from the `slice_colors` array, and the transparency 
      is controlled by the `alpha` parameter.
    - Complex data is plotted using the real part only.
    """

    fig,ax = plt.subplots()
    for idx, slice_idx in enumerate(selected_slices):
        slice_data = np.real(data[slice_idx]) if np.iscomplexobj(data) else data[slice_idx]
        ax.plot(x,slice_data + idx * spacing, color=slice_colors[idx % len(slice_colors)], alpha=0.7)

    return fig,ax

def surf_plot(data,x,y,slice_len,selected_slices):

    """
    Create a 3D surface plot for selected slices of 2D EPR data.

    Parameters
    ----------
    data : ndarray
        The 2D data array from which the slices will be plotted.
    x : ndarray
        The x-axis values corresponding to the data.
    y : ndarray
        The y-axis values corresponding to the data.
    slice_len : int
        The length of each slice along the x-axis.
    selected_slices : list of int
        A list of slice indices to be plotted.

    Returns
    -------
    fig : matplotlib.figure.Figure
        The figure object containing the surface plot.
    ax : matplotlib.axes._axes.Axes
        The axes object of the plot.

    Notes
    -----
    - This function creates a 3D surface plot with `x` and `y` as the axes, and the 
      corresponding data values (`Z`) from the selected slices.
    - Complex data is plotted using the real part only.
    - A color bar is added to the plot to indicate the scale of the surface values.
    """

    fig,ax = plt.subplots(subplot_kw={"projection": "3d"})
    X, Y = np.meshgrid(x[range(slice_len)], y[selected_slices])
    Z = np.real(data[selected_slices, :]) if np.iscomplexobj(data) else data[selected_slices, :]
    surf = ax.plot_surface(X, Y, Z, cmap='jet',rstride=1, cstride=1)
    fig.colorbar(surf, ax=ax, shrink=0.5, aspect=10)

    return fig,ax

def superimposed_plot(data,x,y,selected_slices,slice_colors):

    """
    Create a superimposed plot for selected slices of 2D EPR data.

    Parameters
    ----------
    data : ndarray
        The 2D data array from which the slices will be plotted.
    x : ndarray
        The x-axis values corresponding to the data.
    y : ndarray
        The y-axis values corresponding to the data.
    selected_slices : list of int
        A list of slice indices to be plotted.
    slice_colors : ndarray
        An array of colors to be used for each slice.

    Returns
    -------
    fig : matplotlib.figure.Figure
        The figure object containing the superimposed plot.
    ax : matplotlib.axes._axes.Axes
        The axes object of the plot.

    Notes
    -----
    - This function overlays all selected slices onto a single plot.
    - Complex data is plotted using the real part only.
    - Each slice is plotted with a different color from the `slice_colors` array, 
      and transparency is controlled by the `alpha` parameter.
    - Only the first slice is labeled in the legend, others are excluded using "_nolegend_".
    """

    fig,ax = plt.subplots()
    for idx, slice_idx in enumerate(selected_slices):
        slice_data = np.real(data[slice_idx]) if np.iscomplexobj(data) else data[slice_idx]
        ax.plot(x,slice_data, color=slice_colors[idx % len(slice_colors)], alpha=0.5, label=f'Slice {slice_idx}' if idx == 0 else "_nolegend_")
    
    return fig,ax


def pcolor_plot(data,x,y,slice_len,selected_slices):

    """
    Create a pseudocolor plot for selected slices of 2D EPR data.

    Parameters
    ----------
    data : ndarray
        The 2D data array from which the slices will be plotted.
    x : ndarray
        The x-axis values corresponding to the data.
    y : ndarray
        The y-axis values corresponding to the data.
    slice_len : int
        The length of each slice along the x-axis.
    selected_slices : list of int
        A list of slice indices to be plotted.

    Returns
    -------
    fig : matplotlib.figure.Figure
        The figure object containing the pseudocolor plot.
    ax : matplotlib.axes._axes.Axes
        The axes object of the plot.

    Notes
    -----
    - This function creates a pseudocolor plot (a 2D heatmap) for the selected slices of the data.
    - Complex data is plotted using the real part only.
    - A color bar is added to the plot to indicate the scale of the values.
    """

    fig,ax = plt.subplots()
    X, Y = np.meshgrid(x[range(slice_len)], y[selected_slices])
    Z = np.real(data[selected_slices, :]) if np.iscomplexobj(data) else data[selected_slices, :]
    pc = ax.pcolor(X, Y, Z, cmap='jet')
    fig.colorbar(pc)

    return fig,ax


def slider_plot(data, x, y, slice_len, selected_slices):

    """
    Create a plot for selected slices of 2D EPR data with a slider.

    Parameters
    ----------
    data : ndarray
        The 2D data array from which the slices will be plotted.
    x : ndarray
        The x-axis values corresponding to the data.
    y : ndarray
        The y-axis values corresponding to the data.
    slice_len : int
        The length of each slice along the x-axis.
    selected_slices : list of int
        A list of slice indices to be plotted.

    Returns
    -------
    fig : matplotlib.figure.Figure
        The figure object containing the pseudocolor plot.
    ax : matplotlib.axes._axes.Axes
        The axes object of the plot.

    """

    fig, ax = plt.subplots(figsize=(8, 6))
    plt.subplots_adjust(left=0.15, bottom=0.3)

    X, Y = x[:slice_len], y[selected_slices]
    Z = np.real(data[selected_slices, :]) if np.iscomplexobj(data) else data[selected_slices, :]

    m, n = Z.shape  
    swap_axes = [False]  
    current_x_lim = None 

    # Initial plot (first row)
    line, = ax.plot(X, Z[0, :])
    ax.set_ylim(Z.min(), Z.max())
    ax.autoscale(False)

    # Slider
    ax_slider = plt.axes([0.15, 0.15, 0.7, 0.03])
    slider = Slider(ax_slider, "Index", 0, m - 1, valinit=0, valstep=1,handle_style={'facecolor':'red','size':15},track_color='black')
    fig.slider = slider

    # Swap Axes Button
    ax_button = plt.axes([0.15, 0.07, 0.2, 0.05]) 
    button = Button(ax_button, "Swap Axes")
    fig.button = button 

    # Value label
    ax_label = plt.axes([0.4, 0.07, 0.3, 0.05]) 
    ax_label.set_xticks([]) 
    ax_label.set_yticks([])
    label_text = ax_label.annotate(
        f"{Y[0]}", xy=(0.5, 0.5), ha="center", va="center", fontsize=10
    )

    def update(val):
        """Updates the plot based on slider value and axis mode."""
        index = int(slider.val)
        is_row_mode = not swap_axes[0]  

        if is_row_mode:
            x_data = X 
            y_data = Z[index, :]  
            ax.set_xlim(min(X), max(X)) 
            label_text.set_text(f"{Y[index]}")
        else:
            x_data = Y  
            y_data = Z[:, index]  
            ax.set_xlim(min(Y), max(Y))  
            label_text.set_text(f"{X[index]}")

        if current_x_lim:
            ax.set_xlim(current_x_lim)
        else:
            ax.autoscale(False) 
          
        line.set_xdata(x_data)
        line.set_ydata(y_data)

        fig.canvas.draw_idle()  

    def swap_axes_clicked(event):
        """Swaps row/column mode and updates slider limits."""
        swap_axes[0] = not swap_axes[0]  
        new_max = n - 1 if swap_axes[0] else m - 1

        slider.valmax = new_max
        slider.ax.set_xlim(slider.valmin, slider.valmax)
        slider.set_val(slider.valmin)  
        update(0)  # Force plot update

        if swap_axes[0]:  
            ax.set_xlim(min(Y), max(Y))  
        else:  
            ax.set_xlim(min(X), max(X))  

        fig.canvas.draw_idle()  

    def on_zoom(event):
        """Handles zooming by the user."""
        nonlocal current_x_lim
        current_x_lim = ax.get_xlim() 

    # Attach event listeners
    slider.on_changed(update)
    button.on_clicked(swap_axes_clicked)

    # Add zoom event listener
    fig.canvas.mpl_connect('button_release_event', on_zoom)

    return fig, ax


def interactive_points_selector(x,y):

    """
    Interactively select points from a plot of 1D EPR data.

    Parameters
    ----------
    x : ndarray
        The x-axis values corresponding to the data.
    y : ndarray
        The y-axis values corresponding to the data.

    Returns
    -------
    selected_points : ndarray
        An array of the indices of the selected points, sorted in ascending order.

    Notes
    -----
    - This function creates an interactive plot where points can be selected by left-clicking 
      on the plot. After selecting all desired points, the user can click the 'Done' button 
      to finalize the selection.
    - The function returns the indices of the selected points, sorted in the order of their 
      x-axis values.
    """
    
    fig,ax =plt.subplots()
    ax.plot(x,y)
    ax.set_title('Select points by left click. Click Done after selecting all points.')
    selected_points = []

    def clicked(event):
        if event.inaxes == ax:
            x_id = event.xdata
            idx = int((np.abs(x - x_id)).argmin())
            selected_points.append(idx)
            ax.plot(x[idx],y[idx],'rx')
            fig.canvas.draw()
    def done(event):
        plt.close(fig)
    
    done_button_ax = plt.axes([0.8, 0.05, 0.1, 0.075])
    done_button = Button(done_button_ax, 'Done')
    done_button.on_clicked(done)
    fig.canvas.mpl_connect('button_press_event', clicked)

    # block function until figure is closed.
    plt.show(block=True)
    
    ## sort the points
    selected_points_sorted = sorted(selected_points, key=lambda idx: x[idx])

    return np.unique(np.array(selected_points_sorted, dtype=int))

def data_cursor(fig=None):

    """
    Adds an interactive data cursor to a Matplotlib figure.

    This function enables a crosshair cursor that tracks mouse movement 
    within a given figure and displays the current x and y coordinates.
    Additionally, it supports measuring horizontal distances between 
    two x-coordinates using right-click dragging or Ctrl + Left Click.

    Parameters
    ----------
    fig : matplotlib.figure.Figure, optional
        The Matplotlib figure to which the data cursor will be added. 
        If None, the current active figure (`plt.gcf()`) is used.

    Notes
    -----
    - A red dashed crosshair follows the mouse position.
    - Clicking (Left Click) sets a reference vertical line.
    - Right Click or Ctrl + Left Click enables measuring horizontal distance.
    - The measured distance (ΔX) is displayed in the bottom-left of the figure.
    - The reference vertical line is removed on releasing the mouse button.
    """
    
    if fig is None:
        fig = plt.gcf()

    ax = fig.gca()
    xlim = ax.get_xlim()
    ylim = ax.get_ylim()

    h_line, = ax.plot([xlim[0], xlim[1]], [(ylim[0] + ylim[1]) / 2] * 2, 'r--', lw=1)
    v_line, = ax.plot([(xlim[0] + xlim[1]) / 2] * 2, [ylim[0], ylim[1]], 'r--', lw=1)

    coord_text = ax.text(0.02, 0.95, '', transform=ax.transAxes, fontsize=10, color='red')
    dist_text = ax.text(0.02, 0.02, '', transform=ax.transAxes, fontsize=10, color='blue')
    ref_v_line, = ax.plot([], [], 'b-', lw=1)
    drag_v_line, = ax.plot([], [], 'b-', lw=1)

    ref_x = None
    ctrl_pressed = False  

    def on_mouse_move(event):
        nonlocal ref_x
        if event.inaxes:
            x, y = event.xdata, event.ydata
            h_line.set_ydata([y, y])
            v_line.set_xdata([x, x])
            coord_text.set_text(f'X: {x:.2f}, Y: {y:.2f}')

            # Handling dragging for right-click OR Ctrl + Left Click
            if (event.button == 3 or (ctrl_pressed and event.button is None)):
                if ref_x is None:  
                    ref_x = x 
                    ref_v_line.set_xdata([ref_x, ref_x])
                    ref_v_line.set_ydata(ax.get_ylim())

                dist_text.set_text(f'ΔX: {abs(x - ref_x):.2f}')
                drag_v_line.set_xdata([x, x])
                drag_v_line.set_ydata(ax.get_ylim())
            else:
                dist_text.set_text('')

            fig.canvas.draw_idle()

    def on_mouse_press(event):
        nonlocal ref_x
        if event.button == 1 and event.inaxes and not ctrl_pressed:
            ref_x = event.xdata
            ref_v_line.set_xdata([ref_x, ref_x])
            ref_v_line.set_ydata(ax.get_ylim())

    def on_mouse_release(event):
        nonlocal ref_x
        if event.button == 1 or not ctrl_pressed:
            ref_x = None
            ref_v_line.set_xdata([])
            ref_v_line.set_ydata([])
            drag_v_line.set_xdata([])
            drag_v_line.set_ydata([])
            dist_text.set_text('')
            fig.canvas.draw_idle()

    def on_key_press(event):
        nonlocal ctrl_pressed
        if event.key == 'control':
            ctrl_pressed = True

    def on_key_release(event):
        nonlocal ctrl_pressed
        if event.key == 'control':
            ctrl_pressed = False

    fig.canvas.mpl_connect('motion_notify_event', on_mouse_move)
    fig.canvas.mpl_connect('button_press_event', on_mouse_press)
    fig.canvas.mpl_connect('button_release_event', on_mouse_release)
    fig.canvas.mpl_connect('key_press_event', on_key_press)
    fig.canvas.mpl_connect('key_release_event', on_key_release)

    plt.show()


from pathlib import Path
import sys
import pytest
import numpy as np

from eprpy.loader import check_filepaths,read_DSC_file,get_dim_arrays,get_DTA_datatype,read_DTA_file,read_GF_file
import eprpy as epr

current_dir = Path(__file__).resolve().parent
data_dir = str(current_dir / 'data')
python_version = sys.version_info  

@pytest.fixture
def test_load_1d():
    epr_1d_data = epr.load(data_dir+'/tempo.DSC')
    return epr_1d_data

@pytest.fixture
def test_load_2d():
    epr_2d_data = epr.load(data_dir+'/tempo_time.DSC')
    return epr_2d_data

def test_check_filepaths():
    test_filepath_dsc = data_dir+'/tempo.DSC'
    test_filepath_dta = data_dir+'/tempo.DTA'


    dta_filepath,dsc_filepath = check_filepaths(test_filepath_dsc)
    dta_filepath1,dsc_filepath1 = check_filepaths(test_filepath_dta)

    assert dta_filepath.parts[-1] == 'tempo.DTA'
    assert dsc_filepath.parts[-1] == 'tempo.DSC'
    assert dta_filepath1.parts[-1] == 'tempo.DTA'
    assert dsc_filepath1.parts[-1] == 'tempo.DSC'

    with pytest.raises(FileNotFoundError):
        test_filepath_dsc = str(data_dir)+'/tempo1.DSC'
        dta_filepath,dsc_filepath = check_filepaths(test_filepath_dsc)
    with pytest.raises(FileNotFoundError):
        test_filepath_dta = str(data_dir)+'/tempo1.DTA'
        dta_filepath,dsc_filepath = check_filepaths(test_filepath_dta)

def test_read_DSC_file():
    test_filepath_dsc = data_dir+'/tempo.DSC'

    exp_dict = {'DSRC': 'EXP','BSEQ': 'BIG','IKKF': 'REAL','XTYP': 'IDX','YTYP': 'NODATA','ZTYP': 'NODATA','IRFMT': 'D',
    'XPTS': 2048,'XMIN': 3259.75,'XWID': 130.136426,'TITL': 'tempo','IRNAM': 'Intensity','XNAM': 'Field','IRUNI': '','XUNI': 'G',
    'OPER': 'xuser','DATE': '12/03/20','TIME': '10:57:45','STAG': 'C','EXPT': 'CW','OXS1': 'IADC','AXS1': 'B0VL','AXS2': 'NONE',
    'A1CT': 0.332485,'A1SW': 0.01302,'MWFQ': '9.327654e+09','MWPW': 0.002,'AVGS': 109,'SPTP': 0.03,'RCAG': 60,'RCHM': 1,'B0MA': '4e-05',
    'B0MF': 100000,'RCPH': 0.0,'RCOF': 0.0,'A1RS': 2048,'RCTC': 0,'AllegroMode': 'True','CenterField': '3324.85 G','Delay': '0.0 s','FieldFlyback': 'On',
    'FieldWait': 'Wait LED off','GFactor': 2.0,'MeasuringHall': 'False','SetToSampleG': 'False','StaticFieldMon': '3480.000 G','SweepDirection': 'Up',
    'SweepWidth': '130.2 G','WidthTM': '200.0 G','FrequencyMon': '9.327654 GHz','QMonitBridge': 'On','AcqFineTuning': 'Never','AcqScanFTuning': 'Off',
    'AcqSliceFTuning': 'Off','BridgeCalib': 60.0,'Power': '2.000 mW','PowerAtten': '20.0 dB','BaselineCorr': 'Off','NbScansAcc': 109,'NbScansDone': 109,
    'NbScansToDo': 200,'ReplaceMode': 'Off','SmoothMode': 'Manual','SmoothPoints': 1,'AFCTrap': 'True','AllowShortCt': 'False','Calibrated': 'True','ConvTime': '30.00 ms',
    'DModDetectSCT': 'First','DoubleModAcc': 1,'DoubleModFreq': '5.000 kHz','DoubleMode': 'False','DualDetect': 'OFF','EliDelay': '1.0 us',
    'Enable1stHarm': 'True','Enable1stHarm90': 'False','Enable2ndHarm': 'False','Enable2ndHarm90': 'False','Enable3rdHarm': 'False','Enable3rdHarm90': 'False',
    'Enable4thHarm': 'False','Enable4thHarm90': 'False','Enable5thHarm': 'False','Enable5thHarm90': 'False','EnableDisp': 'False','EnableImag': 'Disable','ExtLockIn': 'False',
    'ExtTrigger': 'False','Gain': '60 dB','GainB': '60 dB','Harmonic': 1,'HighPass': 'True','InputPlugA': 'AC3','InputPlugB': 'AC3','Integrator': 'False','IsCalibExp': 'False',
    'ModAmp': '0.400 G','ModFreq': '100.00 kHz','ModPhase': 0.0,'Offset': '0.0 %','QuadMode': 'False','Resolution': 2048,'Resonator': 1,'SctNorm': 'True',
    'SctRevision': 'Allegro','SetAllOrd': 'False','SetOrdDef': 'False','SpuExtension': 'True','SpuRevision': 'MultiHarmonic',
    'SweepTime': '61.44000 s','TimeConst': 0,'TimeExp': 'False','TuneCaps': 45,'dModSW': 'True'}
    
    assert exp_dict==read_DSC_file(Path(test_filepath_dsc))


def test_get_dim_arrays():
    test_filepath_dsc = data_dir+'/tempo.DSC'
    test_filepath_dta = data_dir+'/tempo.DTA'
    test_dsc_dict = read_DSC_file(Path(test_filepath_dsc))
    exp_dim = np.load(data_dir+'/test_get_dim_array.npy')

    test_ntup,test_dim = get_dim_arrays(test_dsc_dict,Path(test_filepath_dta))
    
    assert test_ntup == (1, 1, 2048)
    np.testing.assert_array_almost_equal(exp_dim,test_dim)

def test_get_DTA_datatype():
    test_filepath_dsc = data_dir+'/tempo.DSC'
    test_dsc_dict = read_DSC_file(Path(test_filepath_dsc))

    assert get_DTA_datatype(test_dsc_dict) == (np.dtype('>f8'), False)

def test_read_DTA_file():
    test_filepath_dsc = data_dir+'/tempo.DSC'
    test_filepath_dta = data_dir+'/tempo.DTA'
    test_dsc_dict = read_DSC_file(Path(test_filepath_dsc))

    exp_datadata = np.load(data_dir+'/test_read_dta_file_data.npy')
    exp_dimlist0 = np.load(data_dir+'/test_read_dta_file_dimlist0.npy')

    test_datadta,test_dimlist = read_DTA_file(Path(test_filepath_dta),test_dsc_dict)
    
    np.testing.assert_array_almost_equal(exp_datadata,test_datadta)
    np.testing.assert_array_almost_equal(test_dimlist[0],exp_dimlist0)
    assert isinstance(test_dimlist,list)

def test_read_GF_file():
    test_filepath_dsc = data_dir+'/tempo.DSC'
    test_filepath_dta = data_dir+'/tempo.DTA'
    test_dsc_dict = read_DSC_file(Path(test_filepath_dsc))

    with pytest.raises(FileNotFoundError):
        _ = read_GF_file('Y',test_dsc_dict,Path(test_filepath_dta))
    

    test_filepath_dsc = data_dir+'/tempo_time.DSC'
    test_filepath_dta = data_dir+'/tempo_time.DTA'
    test_dsc_dict = read_DSC_file(Path(test_filepath_dsc))
    exp_gf_data = np.load(data_dir+'/test_read_GF_file_data.npy')

    np.testing.assert_array_almost_equal(exp_gf_data,read_GF_file('Y',test_dsc_dict,Path(test_filepath_dta)))

    with pytest.raises(FileNotFoundError):
        _ = read_GF_file('X',test_dsc_dict,Path(test_filepath_dta))

def test_load_error():
    with pytest.raises(FileNotFoundError):
        _ = epr.load(data_dir+'/tempo1.DSC')

class TestEPR1d():
    def test_load_1d_epr(self,test_load_1d):

        assert type(test_load_1d) is epr.loader.EprData
        assert test_load_1d.y is None

        exp_x = np.load(data_dir+'/test_load_1d_epr_x.npy')
        exp_data = np.load(data_dir+'/test_load_1d_epr.npy')

        np.testing.assert_array_almost_equal(exp_x,test_load_1d.x)
        np.testing.assert_array_almost_equal(exp_data,test_load_1d.data)

        test_filepath_dsc = data_dir+'/tempo.DSC'
        test_dsc_dict = read_DSC_file(Path(test_filepath_dsc))

        assert test_dsc_dict == test_load_1d.acq_param
        assert test_load_1d.is_complex is False
        assert test_load_1d.filepath == test_filepath_dsc

    def test_scale_between(self,test_load_1d):

        exp_data = np.load(data_dir+'/test_scale_between1d_default.npy')
        np.testing.assert_array_almost_equal(test_load_1d.scale_between().data,exp_data)

        exp_data = np.load(data_dir+'/test_scale_between1d-1_1.npy')
        np.testing.assert_array_almost_equal(test_load_1d.scale_between(-1,1).data,exp_data)

    def test_integral(self,test_load_1d):

        exp_data = np.load(data_dir+'/test_integral.npy')
        np.testing.assert_array_almost_equal(test_load_1d.integral().data,exp_data)

    def test_baseline_correction(self,test_load_1d):

        exp_data = np.load(data_dir+'/test_default_bc.npy')
        np.testing.assert_array_almost_equal(test_load_1d.baseline_correct().data,exp_data)

        exp_data = np.load(data_dir+'/test_poly2_bc.npy')
        np.testing.assert_array_almost_equal(test_load_1d.baseline_correct(method='polynomial',order=2).data,exp_data)

        exp_data = np.load(data_dir+'/test_poly2_bc_baseline.npy')
        np.testing.assert_array_almost_equal(test_load_1d.baseline_correct(method='polynomial',order=2).baseline,exp_data)

        exp_data = np.load(data_dir+'/test_spline_bc.npy')
        np.testing.assert_array_almost_equal(test_load_1d.baseline_correct(method='spline').data,exp_data)

        exp_data = np.load(data_dir+'/test_spline_bc_baseline.npy')
        np.testing.assert_array_almost_equal(test_load_1d.baseline_correct(method='spline').baseline,exp_data)

    def test_plot(self,test_load_1d):
        fig,ax=test_load_1d.plot()
        fig.savefig(data_dir+'/test_results/1dplot.png')

class TestEPR2d():

    def test_load_2d_epr(self,test_load_2d):

        assert type(test_load_2d) is epr.loader.EprData
        assert test_load_2d.y is not None

        exp_x = np.load(data_dir+'/test_load_2d_epr_x.npy')
        exp_data = np.load(data_dir+'/test_load_2d_epr.npy')
        exp_y = np.load(data_dir+'/test_load_2d_epr_y.npy')

        np.testing.assert_array_almost_equal(exp_x,test_load_2d.x)
        np.testing.assert_array_almost_equal(exp_y,test_load_2d.y)
        np.testing.assert_array_almost_equal(exp_data,test_load_2d.data)

        test_filepath_dsc = data_dir+'/tempo_time.DSC'
        test_dsc_dict = read_DSC_file(Path(test_filepath_dsc))

        assert test_dsc_dict == test_load_2d.acq_param
        assert test_load_2d.is_complex is False
        assert test_load_2d.filepath == test_filepath_dsc

    def test_scale_between(self,test_load_2d):

        exp_data = np.load(data_dir+'/test_scale_between2d_default.npy')
        np.testing.assert_array_almost_equal(test_load_2d.scale_between().data,exp_data)

    def test_baseline_correction(self,test_load_2d):

        exp_data = np.load(data_dir+'/test_default2d_bc.npy')
        np.testing.assert_array_almost_equal(test_load_2d.baseline_correct().data,exp_data)

    def test_plot(self,test_load_2d):
        fig,ax=test_load_2d.plot()
        fig.savefig(data_dir+'/test_results/2dplot.png')

        fig,ax=test_load_2d.plot(slices=range(0,20))
        fig.savefig(data_dir+'/test_results/2dplot_range.png')
        
        fig,ax=test_load_2d.plot(plot_type='superimposed')
        fig.savefig(data_dir+'/test_results/2dplot_superimposed.png')
        
        fig,ax=test_load_2d.plot(plot_type='pcolor')
        fig.savefig(data_dir+'/test_results/2dplot_pcolor.png')
        
        fig,ax=test_load_2d.plot(plot_type='surf')
        fig.savefig(data_dir+'/test_results/2dplot_surf.png')


import numpy as np
from datetime import datetime
from copy import deepcopy
from scipy.interpolate import UnivariateSpline

# EPRpy
from eprpy.plotter import interactive_points_selector

def _scale_between(eprdata,min_val=None,max_val=None):

    """
    Scale the data in an `EprData` object to a specified range.

    This function normalizes the data of the given `EprData` object to the range 
    defined by `min_val` and `max_val`. If these values are not provided, the 
    default range is [0, 1].

    Parameters
    ----------
    eprdata : EprData
        An instance of the `EprData` class containing the data to be scaled.
    min_val : float, optional
        The minimum value of the desired range. Defaults to 0.
    max_val : float, optional
        The maximum value of the desired range. Defaults to 1.

    Notes
    -----
    - The function modifies the `data` attribute of the input `EprData` object in place.
    - The scaling operation is performed along the last axis of the `data`.
    - A history entry is added to the `EprData` object, documenting the scaling 
      operation and the range used.
    """
    eprdata_proc = deepcopy(eprdata)

    min_val = 0 if min_val is None else min_val
    max_val = 1 if max_val is None else max_val

    data_min = np.min(eprdata_proc.data,axis=-1,keepdims=True)
    data_max = np.max(eprdata_proc.data,axis=-1,keepdims=True)

    eprdata_proc.data = (eprdata_proc.data-data_min)/(data_max-data_min)
    eprdata_proc.data *= int(max_val-min_val)
    eprdata_proc.data += min_val

    eprdata_proc.history.append([f'{str(datetime.now())} : Data scaled between {min_val} and {max_val}.',deepcopy(eprdata_proc)])

    return eprdata_proc


def _integrate(eprdata):
    
    """
    Compute the integral of the data in an `EprData` object.

    This function calculates the cumulative integral of the data in the 
    `EprData` object along the last axis. The integration step size is determined by the spacing of `eprdata.x`.

    Parameters
    ----------
    eprdata : EprData
        An instance of the `EprData` class containing the data to be integrated.

    Notes
    -----
    - The function modifies the `data` attribute of the input `EprData` object in place.
    - The integration is performed along the last axis of the `data`.
    - A history entry is added to the `EprData` object, documenting the integration 
      operation.
    """
    eprdata_proc = deepcopy(eprdata)

    delta_B = np.mean(np.diff(eprdata_proc.x))
    integral = np.cumsum(eprdata_proc.data,axis=-1)*delta_B 

    eprdata_proc.data = integral
    eprdata_proc.history.append([f'{str(datetime.now())} : Integral calculated',deepcopy(eprdata_proc)])

    return eprdata_proc

def _baseline_correct(eprdata,interactive=False,
                      npts=10,method='linear',spline_smooth=1e-5,
                      order=2):

    """
    Perform baseline correction on the data in an `EprData` object.

    This function removes the baseline from the data in the `EprData` object. 
    It supports both 1D and 2D data, with optional interactive selection of 
    baseline points for 1D data. The baseline can be fitted using linear, 
    polynomial, or spline methods.

    Parameters
    ----------
    eprdata : EprData
        An instance of the `EprData` class containing the data to be baseline-corrected.
    interactive : bool, optional
        If `True`, enables interactive selection of baseline points for 1D data.
        Defaults to `False`.
    npts : int, optional
        Number of points to use from the start and end of the data for baseline fitting
        when `interactive` is `False`. Defaults to 10.
    method : {'linear', 'polynomial', 'spline'}, optional
        The method used for baseline fitting. Choices are:
        - 'linear': Fits a linear baseline.
        - 'polynomial': Fits a polynomial baseline.
        - 'spline': Fits a spline baseline. 
        Defaults to 'linear'.
    spline_smooth : float, optional
        Smoothing factor for spline fitting. Ignored if `method` is not 'spline'. 
        Defaults to 1e-5.
    order : int, optional
        The order of the polynomial for baseline fitting when `method` is 'polynomial'. 
        Defaults to 2.

    Notes
    -----
    - For 2D data, baseline correction is performed using a separate helper function.
    - For 1D data, baseline points can be selected interactively or automatically 
      based on the specified number of points (`npts`).
    - The function modifies the `data` attribute of the input `EprData` object in place.
    - The computed baseline is stored in the `baseline` attribute of the `EprData` object.
    - A history entry is added to the `EprData` object, documenting the baseline correction.
    """

    eprdata_proc = deepcopy(eprdata)

    x = eprdata_proc.x
    y = eprdata_proc.data

    if y.ndim ==2:
        bc_data,baselines = _baseline_correct_2d(x,y,interactive,
                          npts,method,spline_smooth,order)
        eprdata_proc.data = bc_data
        eprdata_proc.baseline = baselines
        eprdata_proc.history.append([f'{str(datetime.now())} : Baseline corrected',deepcopy(eprdata_proc)])

    elif y.ndim==1:

        if interactive:
            baseline_points = interactive_points_selector(x,y)
        else:
            baseline_points=np.concatenate([np.arange(npts), np.arange(len(y) - npts, len(y))])

        # Use the specified baseline points for fitting
        if baseline_points is not None and len(baseline_points) > 0:
            x_fit = x[baseline_points]
            y_fit = y[baseline_points]
        else:
            # If no baseline points are selected, raise an error or use a default (e.g., endpoints)
            raise ValueError("No baseline points selected. Please select points for baseline correction.")

        # Baseline fitting based on selected method
        if method == "linear":
            coeffs = np.polyfit(x_fit, y_fit, 1)
            baseline = np.polyval(coeffs, x)
        elif method == "polynomial":
            coeffs = np.polyfit(x_fit, y_fit, order)
            baseline = np.polyval(coeffs, x)
        elif method == "spline":
            spline = UnivariateSpline(x_fit, y_fit, s=spline_smooth)
            baseline = spline(x)
        else:
            raise ValueError("Method must be 'linear', 'polynomial', or 'spline'.")

        bc_corr = y - baseline

        eprdata_proc.baseline = baseline
        eprdata_proc.data = bc_corr
        eprdata_proc.history.append([f'{str(datetime.now())} : Baseline corrected',deepcopy(eprdata_proc)])

    return eprdata_proc


def _baseline_correct_2d(x,y,interactive=False,
                      npts=10,method='linear',spline_smooth=1e-5,
                      order=2):

    """
    Perform baseline correction on 2D data.

    This function removes the baseline from 2D data, where each row of the input 
    represents a separate spectrum or data series. The baseline can be fitted 
    using linear, polynomial, or spline methods. Optionally, baseline points 
    can be selected interactively.

    Parameters
    ----------
    x : array-like
        The x-axis values corresponding to the data.
    y : array-like
        The 2D data array where each row corresponds to a separate data series.
    interactive : bool, optional
        If `True`, enables interactive selection of baseline points for the 
        first row of the data. Defaults to `False`.
    npts : int, optional
        Number of points to use from the start and end of each row for baseline 
        fitting when `interactive` is `False`. Defaults to 10.
    method : {'linear', 'polynomial', 'spline'}, optional
        The method used for baseline fitting. Choices are:
        - 'linear': Fits a linear baseline.
        - 'polynomial': Fits a polynomial baseline.
        - 'spline': Fits a spline baseline. 
        Defaults to 'linear'.
    spline_smooth : float, optional
        Smoothing factor for spline fitting. Ignored if `method` is not 'spline'. 
        Defaults to 1e-5.
    order : int, optional
        The order of the polynomial for baseline fitting when `method` is 'polynomial'. 
        Defaults to 2.
    
    Returns
    -------
    corrected_data : ndarray
        The 2D data array with baselines removed.
    baselines : ndarray
        The 2D array of computed baselines for each row.

    Notes
    -----
    - Baseline points can be selected interactively for the first row, and the same 
      points are used for all rows.
    - The function supports linear, polynomial, and spline-based baseline fitting.
    - The returned `corrected_data` has the baseline removed, while `baselines` 
      contains the computed baseline for each row of data.
    """

    if interactive:
        baseline_points = interactive_points_selector(x,y[0])
    else:
        baseline_points=np.concatenate([np.arange(npts), np.arange(len(y[0]) - npts, len(y[0]))])
    
    baselines = np.empty_like(y)
    if baseline_points is not None and len(baseline_points) > 0:
        x_fit = x[baseline_points]
        
    for idx,arr in enumerate(y):
        y_fit = arr[baseline_points]
        if method == "linear":
            coeffs = np.polyfit(x_fit, y_fit, 1)
            baseline = np.polyval(coeffs, x)
        elif method == "polynomial":
            coeffs = np.polyfit(x_fit, y_fit, order)
            baseline = np.polyval(coeffs, x)
        elif method == "spline":
            spline = UnivariateSpline(x_fit, y_fit, s=spline_smooth)
            baseline = spline(x)
        else:
            raise ValueError("Method must be 'linear', 'polynomial', or 'spline'.")

        baselines[idx] = baseline
    
    return y-baselines,baselines
